@@ -27,6 +27,7 @@ def create_precomputed(
     green_threads:bool=False, use_https:bool=False,
     max_redirects:int=10, mesh_dir:Optional[str]=None, skel_dir:Optional[str]=None,
     secrets:SecretsType=None, spatial_index_db:Optional[str]=None, 
+    lru_bytes:int = 0,
     **kwargs # absorb graphene arguments
   ):
     path = strict_extract(cloudpath)
@@ -94,6 +95,7 @@ def create_precomputed(
       delete_black_uploads=bool(delete_black_uploads),
       background_color=background_color,
       readonly=readonly,
+      lru_bytes=lru_bytes,
     )
 
     mesh = PrecomputedMeshSource(meta, cache_service, config, readonly)
