@@ -2,7 +2,6 @@ from functools import partial
 import os
 
 import numpy as np
-from six.moves import range
 from tqdm import tqdm
 
 from cloudfiles import CloudFiles, reset_connection_pools, compression
@@ -21,17 +20,19 @@ import cloudvolume.sharedmemory as shm
 
 from ... import check_grid_aligned
 from .common import (
-  fs_lock, parallel_execution, chunknames, 
-  shade
+  parallel_execution, chunknames, shade
 ) 
 from ..common import (
   content_type, cdn_cache_control,
   should_compress
 )
-from .rx import download_chunks_threaded
+from .rx import download_chunks_threaded, decode
+
+progress_queue = None # defined in common.initialize_synchronization
+fs_lock = None # defined in common.initialize_synchronization
 
 def upload(
-    meta, cache,
+    meta, cache, lru,
     image, offset, mip,
     compress=None,
     compress_level=None,
@@ -82,30 +83,38 @@ def upload(
     "secrets": secrets,  
   }
 
+  expanded = bounds.expand_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
+  all_chunks = lambda: set(chunknames(expanded, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip), protocol=meta.path.protocol))
+
+  if parallel > 1:
+    all_chunks = all_chunks()
+    for chunk in all_chunks:
+      lru.pop(chunk, None)
+
   if is_aligned:
     upload_aligned(
-      meta, cache, 
+      meta, cache, lru,
       image, offset, mip,
       **options
     )
     return
 
   # Upload the aligned core
-  expanded = bounds.expand_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
   retracted = bounds.shrink_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
   core_bbox = retracted.clone() - bounds.minpt
 
   if not core_bbox.subvoxel():
     core_img = image[ core_bbox.to_slices() ] 
     upload_aligned(
-      meta, cache, 
+      meta, cache, lru,
       core_img, retracted.minpt, mip,
       **options
     )
 
   # Download the shell, paint, and upload
-  all_chunks = set(chunknames(expanded, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip)))
-  core_chunks = set(chunknames(retracted, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip)))
+  if parallel == 1:
+    all_chunks = all_chunks()
+  core_chunks = set(chunknames(retracted, meta.bounds(mip), meta.key(mip), meta.chunk_size(mip), protocol=meta.path.protocol))
   shell_chunks = all_chunks.difference(core_chunks)
 
   def shade_and_upload(img3d, bbox):
@@ -114,7 +123,7 @@ def upload(
     img3d.setflags(write=1) 
     shade(img3d, bbox, image, bounds)
     threaded_upload_chunks(
-      meta, cache, 
+      meta, cache, lru,
       img3d, mip,
       (( Vec(0,0,0), Vec(*img3d.shape[:3]), bbox.minpt, bbox.maxpt),), 
       compress=compress, cdn_cache=cdn_cache,
@@ -126,7 +135,8 @@ def upload(
   compress_cache = should_compress(meta.encoding(mip), compress, cache, iscache=True)
 
   download_chunks_threaded(
-    meta, cache, mip, shell_chunks, fn=shade_and_upload,
+    meta, cache, None, mip, shell_chunks, 
+    fn=shade_and_upload, decode_fn=decode,
     fill_missing=fill_missing, 
     progress=("Shading Border" if progress else None), 
     compress_cache=compress_cache,
@@ -134,7 +144,7 @@ def upload(
   )
 
 def upload_aligned(
-    meta, cache,
+    meta, cache, lru,
     img, offset, mip,
     compress=None,
     compress_level=None,
@@ -153,11 +163,11 @@ def upload_aligned(
   ):
   global fs_lock
 
-  chunk_ranges = list(generate_chunks(meta, img, offset, mip))
+  chunk_ranges = generate_chunks(meta, img, offset, mip)
 
   if parallel == 1:
     threaded_upload_chunks(
-      meta, cache, 
+      meta, cache, lru,
       img, mip, chunk_ranges, 
       progress=progress,
       compress=compress, cdn_cache=cdn_cache,
@@ -168,13 +178,6 @@ def upload_aligned(
     )
     return
 
-  length = (len(chunk_ranges) // parallel) or 1
-  chunk_ranges_by_process = []
-  for i in range(0, len(chunk_ranges), length):
-    chunk_ranges_by_process.append(
-      chunk_ranges[i:i+length]
-    )
-
   # use_shared_memory means use a predetermined
   # shared memory location, not no shared memory 
   # at all.
@@ -196,7 +199,11 @@ def upload_aligned(
     secrets=secrets
   )
 
-  parallel_execution(cup, chunk_ranges_by_process, parallel, cleanup_shm=location)
+  parallel_execution(
+    cup, chunk_ranges, parallel, 
+    progress, desc="Upload", 
+    cleanup_shm=location
+  )
 
   # If manual mode is enabled, it's the 
   # responsibilty of the user to clean up
@@ -229,24 +236,35 @@ def child_upload_process(
     readonly=True
   )
 
-  if location_bbox:
-    cutout_bbox = Bbox( offset, offset + img_shape[:3] )
-    delta_box = cutout_bbox.clone() - location_bbox.minpt
-    renderbuffer = renderbuffer[ delta_box.to_slices() ]
+  def updatefn():
+    if progress:
+      # This is not good programming practice, but
+      # I could not find a clean way to do this that
+      # did not result in warnings about leaked semaphores.
+      # progress_queue is created in common.py:initialize_progress_queue
+      # as a global for this module.
+      progress_queue.put(1)
 
-  threaded_upload_chunks(
-    meta, cache, 
-    renderbuffer, mip, chunk_ranges, 
-    compress=compress, cdn_cache=cdn_cache, progress=progress,
-    delete_black_uploads=delete_black_uploads, 
-    background_color=background_color,
-    green=green, compress_level=compress_level,
-    secrets=secrets
-  )
-  array_like.close()
+  try: 
+    if location_bbox:
+      cutout_bbox = Bbox( offset, offset + img_shape[:3] )
+      delta_box = cutout_bbox.clone() - location_bbox.minpt
+      renderbuffer = renderbuffer[ delta_box.to_slices() ]
+
+    return threaded_upload_chunks(
+      meta, cache, None,
+      renderbuffer, mip, chunk_ranges, 
+      compress=compress, cdn_cache=cdn_cache, progress=updatefn,
+      delete_black_uploads=delete_black_uploads, 
+      background_color=background_color,
+      green=green, compress_level=compress_level,
+      secrets=secrets,
+    )
+  finally:
+    array_like.close()
 
 def threaded_upload_chunks(
-    meta, cache, 
+    meta, cache, lru,
     img, mip, chunk_ranges, 
     compress, cdn_cache, progress,
     n_threads=DEFAULT_THREADS,
@@ -267,13 +285,20 @@ def threaded_upload_chunks(
   local = CloudFiles('file://' + cache.path, secrets=secrets)
 
   def do_upload(imgchunk, cloudpath):
-    encoded = chunks.encode(imgchunk, meta.encoding(mip), meta.compressed_segmentation_block_size(mip))
+    encoded = chunks.encode(
+      imgchunk, meta.encoding(mip), 
+      meta.compressed_segmentation_block_size(mip),
+      compression_params=meta.compression_params(mip),
+    )
+
+    if lru is not None:
+      lru[cloudpath] = encoded
 
     remote_compress = should_compress(meta.encoding(mip), compress, cache)
     cache_compress = should_compress(meta.encoding(mip), compress, cache, iscache=True)
     remote_compress = compression.normalize_encoding(remote_compress)
     cache_compress = compression.normalize_encoding(cache_compress)
-
+    
     encoded = compression.compress(encoded, remote_compress)
     cache_encoded = encoded
     if remote_compress != cache_compress:
@@ -331,14 +356,22 @@ def threaded_upload_chunks(
     else:
       do_upload(imgchunk, cloudpath)
 
+  def process_and_update(*args, **kwargs):
+    process(*args, **kwargs)
+    # Needed for multiprocess progress bar.
+    if callable(progress):
+      progress()
+
   schedule_jobs(
-    fns=( partial(process, *vals) for vals in chunk_ranges ), 
+    fns=( partial(process_and_update, *vals) for vals in chunk_ranges ), 
     concurrency=n_threads, 
-    progress=('Uploading' if progress else None),
+    progress=('Uploading' if progress and not callable(progress) else None),
     total=len(chunk_ranges),
     green=green,
   )
 
+  return len(chunk_ranges)
+
 def generate_chunks(meta, img, offset, mip):
   shape = Vec(*img.shape)[:3]
   offset = Vec(*offset)[:3]
@@ -348,24 +381,34 @@ def generate_chunks(meta, img, offset, mip):
   alignment_check = bounds.round_to_chunk_size(meta.chunk_size(mip), meta.voxel_offset(mip))
 
   if not np.all(alignment_check.minpt == bounds.minpt):
-    raise AlignmentError("""
+    raise AlignmentError(f"""
       Only chunk aligned writes are supported by this function. 
 
-      Got:             {}
-      Volume Offset:   {} 
-      Nearest Aligned: {}
-    """.format(
-      bounds, meta.voxel_offset(mip), alignment_check)
-    )
+      Got:             {bounds}
+      Volume Offset:   {meta.voxel_offset(mip)} 
+      Nearest Aligned: {alignment_check}
+    """)
 
   bounds = Bbox.clamp(bounds, meta.bounds(mip))
 
   img_offset = bounds.minpt - offset
   img_end = Vec.clamp(bounds.size3() + img_offset, Vec(0,0,0), shape)
 
-  for startpt in xyzrange( img_offset, img_end, meta.chunk_size(mip) ):
-    startpt = startpt.clone()
-    endpt = min2(startpt + meta.chunk_size(mip), shape)
-    spt = (startpt + bounds.minpt).astype(int)
-    ept = (endpt + bounds.minpt).astype(int)
-    yield (startpt, endpt, spt, ept)
+  class ChunkIterator():
+    def __len__(self):
+      csize = meta.chunk_size(mip)
+      bbox = Bbox(img_offset, img_end)
+      # round up and avoid conversion to float
+      n_chunks = (bbox.dx + csize[0] - 1) // csize[0]
+      n_chunks *= (bbox.dy + csize[1] - 1) // csize[1]
+      n_chunks *= (bbox.dz + csize[2] - 1) // csize[2]
+      return n_chunks
+    def __iter__(self):
+      for startpt in xyzrange( img_offset, img_end, meta.chunk_size(mip) ):
+        startpt = startpt.clone()
+        endpt = min2(startpt + meta.chunk_size(mip), shape)
+        spt = (startpt + bounds.minpt).astype(int)
+        ept = (endpt + bounds.minpt).astype(int)
+        yield (startpt, endpt, spt, ept)
+
+  return ChunkIterator()
