@@ -293,9 +293,19 @@ class SpatialIndex(object):
     if parse["scheme"] == "sqlite":
       if parallel != 1:
         raise ValueError("sqlite supports only one writer at a time.")
-      return self.to_sqlite(parse["path"], create_indices, progress, allow_missing)
+      return self.to_sqlite(
+        parse["path"], 
+        create_indices=create_indices, 
+        allow_missing=allow_missing, 
+        progress=progress,
+      )
     elif parse["scheme"] == "mysql":
-      return self.to_mysql(path, create_indices, allow_missing, progress, parallel)
+      return self.to_mysql(path, 
+        create_indices=create_indices, 
+        allow_missing=allow_missing, 
+        progress=progress,
+        parallel=parallel,
+      )
     else:
       raise ValueError(
         f"Unsupported database type. {path}\n"
@@ -585,9 +595,7 @@ def insert_index_files(index_files, lock, conn, cur, progress, mysql_syntax):
   AUTOINC = "AUTO_INCREMENT" if mysql_syntax else "AUTOINCREMENT"
   INTEGER = "BIGINT UNSIGNED" if mysql_syntax else "INTEGER"
 
-  values = ( (os.path.basename(filename),) for filename in index_files.keys() )
-  if mysql_syntax:
-    values = list(values) # doesn't support generators in v8.0.26
+  values = [ os.path.basename(filename) for filename in index_files.keys() ]
 
   # This is a critical region as if multiple inserts and queries occur outside
   # a transaction, you could crash the following code. Fortunately, these 
@@ -596,7 +604,8 @@ def insert_index_files(index_files, lock, conn, cur, progress, mysql_syntax):
   # The select could be rewritten to be order independent at the cost of 
   # sending more data for a (probably) slower query.
   with lock:
-    cur.executemany(f"INSERT INTO index_files(filename) VALUES ({BIND})", values)
+    bindlist = ",".join([f"({BIND})"] * len(values))
+    cur.execute(f"INSERT INTO index_files(filename) VALUES {bindlist}", values)
     cur.execute(f"SELECT filename,id from index_files ORDER BY id desc LIMIT {len(index_files)}")
 
   filename_id_map = { 
@@ -617,11 +626,27 @@ def insert_index_files(index_files, lock, conn, cur, progress, mysql_syntax):
   
   @retry
   def insert_file_lookup_values(cur, chunked_values):
-    cur.executemany(f"INSERT INTO file_lookup(label, fid) VALUES ({BIND},{BIND})", chunked_values)
+    nonlocal BIND
+    bindlist = ",".join([f"({BIND},{BIND})"] * len(chunked_values))
+    flattened_values = []
+    for label, fid in chunked_values:
+      flattened_values.append(label)
+      flattened_values.append(fid)
+    cur.execute(f"INSERT INTO file_lookup(label, fid) VALUES {bindlist}", flattened_values)
 
-  for chunked_values in sip(all_values, 500000):
-    insert_file_lookup_values(cur, chunked_values)
-    conn.commit()
+  pbar = tqdm(
+    desc="Inserting File Lookups", 
+    disable=(not progress), 
+    total=len(all_values)
+  )
+
+  block_size = 500000 if mysql_syntax else 15000
+
+  with pbar:
+    for chunked_values in sip(all_values, block_size):
+      insert_file_lookup_values(cur, chunked_values)
+      conn.commit()
+      pbar.update(len(chunked_values))
 
 def set_journaling_to_performance_mode(cur, mysql_syntax):
   if mysql_syntax:
