@@ -1,3 +1,5 @@
+from typing import Optional, List, Any, Sequence
+
 from collections import defaultdict
 from datetime import datetime
 import math
@@ -13,10 +15,11 @@ import dateutil.parser
 import fastremap
 import numpy as np
 
+from .. import chunks
 from .. import compression
 from .. import exceptions
 from ..cacheservice import CacheService
-from ..lib import Bbox, Vec, toiter
+from ..lib import Bbox, Vec, toiter, BboxLikeType
 from ..storage import SimpleStorage, Storage, reset_connection_pools
 from ..volumecutout import VolumeCutout
 from ..datasource.graphene.metadata import GrapheneApiVersion
@@ -68,7 +71,82 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
   def mesh_chunk_size(self):
     # TODO: add this as new parameter to the info as it can be different from the chunkedgraph chunksize
     return self.meta.mesh_chunk_size
-  
+
+  def scattered_points(
+    self, pts, 
+    mip=None, coord_resolution=None,
+    agglomerate=None, timestamp=None, stop_layer=None,
+  ):
+    """
+    Download one or more single voxel values that may be scattered
+    across the dataset. You can accelerate this query with an LRU
+    if there is some spatial localization.
+
+    pts: iterable of triples
+    mip: which resolution level to get (default self.mip)
+    coord_resolution: (rx,ry,rz) the coordinate resolution of the input point.
+      Sometimes Neuroglancer is working in the resolution of another
+      higher res layer and this can help correct that.
+
+    agglomerate: if true, remap all watershed ids in the volume
+      and return a flat segmentation.
+
+    if agglomerate is true these options are available:
+
+    timestamp: (agglomerate only) get the roots from this date and time
+      formats accepted:
+        int: unix timestamp
+        datetime: self explainatory
+        string: ISO 8601 date
+    stop_layer: (agglomerate only) (int) if specified, return the lowest 
+      parent at or above that layer. If not specified, go all the way 
+      to the root id. 
+        Layer 1: Watershed
+        Layer 2: Within-Chunk Agglomeration
+        Layer 2+: Between chunk interconnections (skip connections possible)
+
+    If agglomerate is None, then the cv.meta.agglomerate controls
+    its value.
+
+    Returns: 
+    
+      { (x,y,z): label, ... }
+    """
+    pts = list(pts)
+    if isinstance(pts[0], int):
+      pts = [ pts ]
+
+    if mip is None:
+      mip = self.mip
+    mip = self.meta.to_mip(mip)
+
+    if coord_resolution is not None:
+      factor = self.meta.resolution(0) / Vec(*coord_resolution)
+      pts = [ Vec(*pt) / factor for pt in pts ]
+
+    pts = set([ tuple(self.point_to_mip(pt, mip=0, to_mip=mip)) for pt in pts ])
+    results = self.image.download_points(pts, mip)
+
+    agglomerate = agglomerate if agglomerate is not None else self.agglomerate
+    if (agglomerate and stop_layer is not None) and (stop_layer <= 0 or stop_layer > self.meta.n_layers):
+      raise ValueError(
+        f"Stop layer {stop_layer} must be "
+        f"1 <= stop_layer <= {self.meta.n_layers} or None."
+      )
+
+    if not agglomerate:
+      return results
+
+    labels = np.array(list(results.values()), dtype=np.uint64)
+    roots = self.agglomerate_cutout(
+      labels, 
+      timestamp=timestamp, 
+      stop_layer=stop_layer,
+      in_place=False,
+    )
+    mapping = { segid: root for segid, root in zip(labels, roots) }
+    return { pt: mapping[label] for pt, label in results.items() }
+
   def download_point(
     self, pt, size=256, 
     mip=None, parallel=None, 
@@ -119,12 +197,215 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
 
     return self.download(bbox, mip, parallel=parallel, **kwargs)
 
+  def unique(
+    self, 
+    bbox:BboxLikeType, 
+    mip:Optional[int] = None, 
+    segids:Optional[List[int]] = None,
+    preserve_zeros:bool = False,
+    agglomerate:Optional[bool] = None, 
+    timestamp:Optional[int] = None,
+    stop_layer:Optional[int] = None,
+    coord_resolution:Optional[Sequence[int]] = None,
+  ) -> set:
+    """
+    Extracts unique labels from segmentation and optionally agglomerates
+    labels based on information in the graph server.
+
+    This operation can be done with download and np.unique but this
+    version scales to much larger sizes and is faster.
+
+    bbox: specifies cutout to fetch
+    mip: which resolution level to get (default self.mip)
+    coord_resolution: (rx,ry,rz) the coordinate resolution of the input point.
+      Sometimes Neuroglancer is working in the resolution of another
+      higher res layer and this can help correct that.
+
+    agglomerate: if true, remap all watershed ids in the volume
+      and return a flat segmentation.
+
+    if agglomerate is true these options are available:
+
+    timestamp: (agglomerate only) get the roots from this date and time
+      formats accepted:
+        int: unix timestamp
+        datetime: self explainatory
+        string: ISO 8601 date
+    stop_layer: (agglomerate only) (int) if specified, return the lowest 
+      parent at or above that layer. If not specified, go all the way 
+      to the root id. 
+        Layer 1: Watershed
+        Layer 2: Within-Chunk Agglomeration
+        Layer 2+: Between chunk interconnections (skip connections possible)
+
+    If agglomerate is None, then the cv.meta.agglomerate controls
+    its value.
+
+    If agglomerate is false, these other options come into play:
+
+    segids: agglomerate the leaves of these segids from the graph 
+      server and label them with the given segid.
+    preserve_zeros: If segids is not None:
+      False: mask other segids with zero
+      True: mask other segids with the largest integer value
+        contained by the image data type and leave zero as is.
+
+    Returns: set of integers
+    """
+    agglomerate = agglomerate if agglomerate is not None else self.agglomerate
+    
+    bbox = Bbox.create(
+      bbox, context=self.bounds, 
+      bounded=(self.bounded and coord_resolution is None), 
+      autocrop=self.autocrop
+    )
+  
+    if mip is None:
+      mip = self.mip
+
+    if coord_resolution is not None:
+      factor = self.meta.resolution(mip) / coord_resolution
+      bbox /= factor
+      if self.bounded and not self.meta.bounds(mip).contains_bbox(bbox):
+        raise exceptions.OutOfBoundsError(f"Computed {bbox} is not contained within bounds {self.meta.bounds(mip)}")
+
+    if bbox.subvoxel():
+      raise exceptions.EmptyRequestException("Requested {} is smaller than a voxel.".format(bbox))
+
+    if (agglomerate and stop_layer is not None) and (stop_layer <= 0 or stop_layer > self.meta.n_layers):
+      raise ValueError(
+        f"Stop layer {stop_layer} must be "
+        f"1 <= stop_layer <= {self.meta.n_layers} or None."
+      )
+
+    mip0_bbox = self.bbox_to_mip(bbox, mip=mip, to_mip=0)
+    # Only ever necessary to make requests within the bounding box
+    # to the server. We can fill black in other situations.
+    mip0_bbox = bbox.intersection(self.meta.bounds(0), mip0_bbox)
+
+    labels = super().unique(bbox, mip=mip)
+
+    if agglomerate:
+      return set(self.get_roots(
+        list(labels), 
+        timestamp=timestamp, 
+        binary=True, 
+        stop_layer=stop_layer
+      ))
+
+    labels = set(labels)
+    if segids is None:
+      return labels
+
+    for segid in segids:
+      leaves = set(self.get_leaves(segid, mip0_bbox, 0))
+      if labels.isdisjoint(leaves):
+        continue
+      labels -= leaves
+      labels.add(segid)
+
+    mask_value = 0
+    if preserve_zeros:
+      mask_value = np.inf
+      if np.issubdtype(self.dtype, np.integer):
+        mask_value = np.iinfo(self.dtype).max
+
+      segids.append(0)
+
+    segids = set(segids)
+    final_labels = set([ label for label in labels if label in segids ])
+    if len(final_labels) < len(labels):
+      final_labels.add(mask_value)
+    return final_labels
+
+  def download_files(
+    self, bbox, mip=None, 
+    parallel=None, segids=None,
+    agglomerate=None, timestamp=None,
+    stop_layer=None,
+    coord_resolution=None,
+    cache_only=False,
+  ):
+    agglomerate = agglomerate if agglomerate is not None else self.agglomerate
+    
+    bbox = Bbox.create(
+      bbox, context=self.bounds, 
+      bounded=(self.bounded and coord_resolution is None), 
+      autocrop=self.autocrop
+    )
+
+    if mip is None:
+      mip = self.mip
+
+    if coord_resolution is not None:
+      factor = self.meta.resolution(mip) / coord_resolution
+      bbox /= factor
+      if self.bounded and not self.meta.bounds(mip).contains_bbox(bbox):
+        raise exceptions.OutOfBoundsError(f"Computed {bbox} is not contained within bounds {self.meta.bounds(mip)}")
+
+    if bbox.subvoxel():
+      raise exceptions.EmptyRequestException(f"Requested {bbox} is smaller than a voxel.")
+
+    if (agglomerate and stop_layer is not None) and (stop_layer <= 0 or stop_layer > self.meta.n_layers):
+      raise ValueError("Stop layer {} must be 1 <= stop_layer <= {} or None.".format(stop_layer, self.meta.n_layers))
+
+    mip0_bbox = self.bbox_to_mip(bbox, mip=mip, to_mip=0)
+    # Only ever necessary to make requests within the bounding box
+    # to the server. We can fill black in other situations.
+    mip0_bbox = bbox.intersection(self.meta.bounds(0), mip0_bbox)
+
+    files = self.image.download_files(
+      bbox, mip=mip, 
+      decompress=True, parallel=parallel,
+      cache_only=cache_only,
+    )
+
+    labels = set([])
+    for file in files.values():
+      labels.update(chunks.labels(
+        file, 
+        encoding=self.meta.encoding(mip),
+        shape=self.meta.chunk_size(mip),
+        dtype=self.meta.dtype,
+        block_size=self.meta.compressed_segmentation_block_size(mip),
+      ))
+
+    def apply_mapping(mapping):
+      for key in files:
+        files[key] = chunks.remap(
+          files[key], 
+          encoding=self.meta.encoding(mip), 
+          shape=self.meta.chunk_size(mip),
+          dtype=self.meta.dtype,
+          block_size=self.meta.compressed_segmentation_block_size(mip),
+          mapping=mapping,
+          preserve_missing_labels=True,
+        )
+      return files
+
+    if agglomerate:
+      labels = list(labels)
+      roots = self.get_roots(labels, timestamp=timestamp, binary=True, stop_layer=stop_layer)
+      return apply_mapping({ segid: root for segid, root in zip(labels, roots) })
+    elif segids is None:
+      return files
+
+    segids = list(toiter(segids))
+
+    remapping = {}
+    for segid in segids:
+      leaves = self.get_leaves(segid, mip0_bbox, 0)
+      remapping.update({ leaf: segid for leaf in leaves })
+    
+    return apply_mapping(remapping)
+
   def download(
     self, bbox, mip=None, 
     parallel=None, segids=None,
     preserve_zeros=False,
     agglomerate=None, timestamp=None,
-    stop_layer=None, renumber=False
+    stop_layer=None, renumber=False,
+    coord_resolution=None,
   ):
     """
     Downloads base segmentation and optionally agglomerates
@@ -133,6 +414,9 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
     bbox: specifies cutout to fetch
     mip: which resolution level to get (default self.mip)
     parallel: what parallel level to use (default self.parallel)
+    coord_resolution: (rx,ry,rz) the coordinate resolution of the input point.
+      Sometimes Neuroglancer is working in the resolution of another
+      higher res layer and this can help correct that.
 
     agglomerate: if true, remap all watershed ids in the volume
       and return a flat segmentation.
@@ -165,26 +449,29 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
 
     Returns: img as a VolumeCutout
     """
-    if type(bbox) is Vec:
-      bbox = Bbox(bbox, bbox+1)
-
     agglomerate = agglomerate if agglomerate is not None else self.agglomerate
     
     bbox = Bbox.create(
       bbox, context=self.bounds, 
-      bounded=self.bounded, 
+      bounded=(self.bounded and coord_resolution is None), 
       autocrop=self.autocrop
     )
-  
+
+    if mip is None:
+      mip = self.mip
+
+    if coord_resolution is not None:
+      factor = self.meta.resolution(mip) / coord_resolution
+      bbox /= factor
+      if self.bounded and not self.meta.bounds(mip).contains_bbox(bbox):
+        raise exceptions.OutOfBoundsError(f"Computed {bbox} is not contained within bounds {self.meta.bounds(mip)}")
+
     if bbox.subvoxel():
       raise exceptions.EmptyRequestException("Requested {} is smaller than a voxel.".format(bbox))
 
     if (agglomerate and stop_layer is not None) and (stop_layer <= 0 or stop_layer > self.meta.n_layers):
       raise ValueError("Stop layer {} must be 1 <= stop_layer <= {} or None.".format(stop_layer, self.meta.n_layers))
 
-    if mip is None:
-      mip = self.mip
-
     mip0_bbox = self.bbox_to_mip(bbox, mip=mip, to_mip=0)
     # Only ever necessary to make requests within the bounding box
     # to the server. We can fill black in other situations.
@@ -233,10 +520,13 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
       return img, renumber_remap
     return img
   
-  def agglomerate_cutout(self, img, timestamp=None, stop_layer=None):
+  def agglomerate_cutout(self, img, timestamp=None, stop_layer=None, in_place=True):
     """Remap a graphene volume to its latest root ids. This creates a flat segmentation."""
     if np.all(img == self.image.background_color) or stop_layer == 1:
-      return img
+      if not in_place:
+        return np.copy(img)
+      else:
+        return img
 
     labels = fastremap.unique(img)
     if labels.size and labels[0] == 0:
@@ -244,9 +534,22 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
 
     roots = self.get_roots(labels, timestamp=timestamp, binary=True, stop_layer=stop_layer)
     mapping = { segid: root for segid, root in zip(labels, roots) }
-    return fastremap.remap(img, mapping, preserve_missing_labels=True, in_place=True)
+    return fastremap.remap(img, mapping, preserve_missing_labels=True, in_place=in_place)
+
+  def coordinate_indexing(self, slices):
+    res = super().coordinate_indexing(slices)
+    if not self.agglomerate:
+      return res
+    return self.agglomerate_cutout(res)
 
   def __getitem__(self, slices):
+    if (
+      hasattr(slices, "__len__") 
+      and len(slices) == 3
+      and all([ isinstance(slc, (list, tuple, np.ndarray)) for slc in slices ])
+    ):
+      return self.coordinate_indexing(slices)
+
     return self.download(
       slices, mip=self.mip,
       preserve_zeros=True,
@@ -264,7 +567,7 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
 
   def get_root(self, segid, *args, **kwargs):
     """Deprecated. Get a single root id for a single segid."""
-    return get_roots(segid, *args, **kwargs)[0]
+    return self.get_roots(segid, *args, **kwargs)[0]
 
   def get_roots(self, segids, timestamp=None, binary=True, stop_layer=None):
     """
@@ -296,7 +599,8 @@ class CloudVolumeGraphene(CloudVolumePrecomputed):
     base_remap = { 0: 0 }
     # skip ids that are already root IDs
     for segid in segids:
-      if self.meta.decode_layer_id(segid) == self.meta.n_layers:
+      layer_id = self.meta.decode_layer_id(segid)
+      if layer_id in (stop_layer, self.meta.n_layers):
         base_remap[segid] = segid
 
     segids = np.array(
