@@ -2,6 +2,7 @@ import os
 import numpy as np
 import pytest 
 
+from cloudfiles import CloudFiles
 from cloudvolume import Vec, Bbox, CloudVolume, Storage, Mesh
 
 @pytest.fixture
@@ -67,6 +68,24 @@ def test_get_mesh(unsharded_vol):
   mesh_orig = vol.mesh.get(94081437)
   assert mesh == mesh_orig
 
+def test_put(unsharded_vol):
+  mesh = Mesh(vertices=[[0,0,0], [1,1,1], [2,2,2]], faces=[0,1,2])
+  mesh.segid = 777
+
+  cf = CloudFiles(unsharded_vol.mesh.meta.layerpath)
+
+  unsharded_vol.mesh.put(mesh)
+  lst = list(cf)
+  assert "777:0" in lst
+  assert "777:0:1" in lst
+  m = unsharded_vol.mesh.get(777, fuse=False)[777]
+  assert len(m.faces) == 1
+  assert m.segid == 777
+  unsharded_vol.mesh.delete(777)
+  lst = list(cf)
+  assert "777:0" not in lst
+  assert "777:0:1" not in lst
+
 def test_duplicate_vertices():
   verts = np.array([
     [0,0,0], [0,1,0],
